module Compiler(compile)
where

import CurryCoreSyntax
import InstructionSet
import CommonSyntax


compile :: CCProgram -> NF_Instructions
---------------------------------------
compile = compileExpr False 0


compileExpr :: Bool -> Int -> CCProgram -> NF_Instructions
----------------------------------------------------------
compileExpr b n (CCExprLet fundefs expr) =
   (NF_LET (length fundefs) (map compileFunDef fundefs)) : (compileExpr b n expr)
compileExpr b n (CCExprFreeDecl _ expr) =
   NF_LETFREE : (compileExpr b n expr)
compileExpr b n (CCExprLambda (CCVariable id) expr) =
   if (b || n > 0) then
      (NF_TAKE id) : (compileExpr b (n-1) expr)
   else
      (NF_VALUE) : (NF_TAKE id) : (compileExpr True n expr)
compileExpr b n (CCExprCase Rigid var alts) =
   (NF_EVAL $ compileVar var):NF_WAIT:(compileSwitch b n alts)
compileExpr b n (CCExprCase Flex var alts) =
   (NF_EVAL $ compileVar var):(NF_BOUND $ compileSwitch b n alts):(compileIndet b n alts)
-- the annotations Choice and DefaultAnnotation can't occur:
compileExpr _ _ (CCExprCase _ _ _) = 
   error "Compiler.compileExpr: Panic! The impossible happened!\n"
compileExpr b n (CCExprAppl expr (CCVariable id)) =
   (NF_ARG id):(compileExpr b (n+1) expr)
compileExpr _ _ (CCExprVar var) =
   compileVar var
compileExpr _ _ (CCExprConstr c vars) =
   compileConstr c vars
compileExpr b n (CCExprStrict expr) =
   (NF_EVAL $ compileExpr b n expr) : [NF_STRICT, NF_RETURN] -- only for "main"
							     -- if we would have type informations,
							     -- strict wouldn't occur if main 
							     -- has functional type
compileExpr b n (CCExprAnd expr1 expr2) =
   (NF_A_THREADS $ compileExpr b n expr1) : (compileExpr b n expr2)
compileExpr b n (CCExprOr expr1 expr2) =
   (NF_TRY $ compileExpr b n expr1) : (NF_TRY $ compileExpr b n expr2) : [NF_BACKTRACK]
compileExpr b n (CCExprUnify expr1 expr2) =
   (NF_U_THREADS $ compileExpr b n expr1) : (compileExpr b n expr2)
compileExpr b n (CCExprGuard expr1 expr2) =
   (NF_EVAL $ compileExpr b n expr1) : compileExpr b n expr2
compileExpr _ _ CCExprFail =
   [NF_BACKTRACK]
compileExpr _ _ (CCExprError error) =
   compileError error

compileFunDef (CCFunDef (CCVariable id) expr) =
   (NF_BLACKHOLE_MARKER id) : (NF_EVAL $ compileExpr False 0 expr) : [NF_UPDATE id, NF_DEREF id] -- DEREF eigentlich nur noetig, wenn expr
                                                                                     -- eine Funktion!
   

compileSwitch b n alts =
   let compileSwitch' b n (CCAlternative (CCDataConstr _ _ _) vars expr) =
          (map (\ _ -> NF_BIND) vars) ++ compileExpr b n expr
   in  [NF_SWITCH $ map (compileSwitch' b n) alts]

compileIndet b n alts = -- with a small optimization!
   NF_WAKE_UP_THREADS : ((map (compileTry b n) [alt | alt@(CCAlternative _ _ e) <- alts, e /= CCExprFail]) ++ [NF_BACKTRACK])


compileTry b n (CCAlternative c vars expr) =
   NF_TRY $ (map (\ _ -> NF_LETFREE) vars) ++ [NF_BIND_TO $ compileConstr c vars] ++ compileExpr b n expr


compileVar (CCVariable id) =
   [NF_DEREF id, NF_UVAR id] 


compileConstr (CCDataConstr (CCName n) id arity) vars =
   if arity == 0 then
      [NF_CONSTANT n id]
    else
      ((NF_CONSTR n id arity $ compilePattern vars):compileStrict vars)

compilePattern vars =
   concatMap compileVar vars

compileStrict []   = error "compileStrict: empty var list!\n"
compileStrict vars = 
   let compileStrictV (CCVariable id) =
            (NF_EVAL [NF_DEREF id, NF_UVAR id]) : [NF_STRICT, NF_RETURN]
   in  (map (\ v -> NF_A_THREADS $ compileStrictV v) (tail vars)) ++ compileStrictV (head vars)


compileError CCErrorPatternDecl = [NF_ERROR "pattern declararation not fitted!\n"] -- ??