module Lexer (Token(..), lexer, parseError, isSymbol) where

import Char(isAlpha, isLower, isUpper, isDigit)
import Maybe(fromMaybe)
import ParseMonad
import CommonSyntax(SrcLoc(..))

data Token 

-- reserved words

   = TokenChoice
   | TokenData
   | TokenElse
   | TokenEval
   | TokenExternal
   | TokenFlex
   | TokenFree
   | TokenIf
   | TokenImport
   | TokenIn
   | TokenInfix
   | TokenInfixl
   | TokenInfixr
   | TokenLet
   | TokenModule
   | TokenOptmatch
   | TokenPragma
   | TokenRenaming
   | TokenRigid
   | TokenThen
   | TokenTo
   | TokenWhere

-- literals

   | TokenIntLiteral String

-- names

   | TokenIdentifier        String
   | TokenDottedIdentifier  [String]
   | TokenConstructor       String
   | TokenDottedConstructor [String]
   | TokenInfixConstructor  String

-- operators/separators

   | TokenArrow
   | TokenEquals
   | TokenBar
   | TokenSig
   | TokenLambda
   | TokenDotDot
   | TokenSemicolon
   | TokenComma
   | TokenAnonymous
   | TokenLParen
   | TokenRParen
   | TokenLBracket
   | TokenRBracket
   | TokenLBrace
   | TokenRBrace
   | TokenVRBrace    -- a virtual close brace

-- infix operators

   | TokenMinus
   | TokenInfixOpId String

   | EOF

-- not used yet

   | TokenError String

reservedOps :: [(String, Token)]
reservedOps  = [("->",        TokenArrow),
                ("=",         TokenEquals),
                ("|",         TokenBar),
                ("::",        TokenSig),
                ("\\",        TokenLambda),
                ("..",        TokenDotDot),
                ("-",         TokenMinus)]

reservedIDs :: [(String, Token)]
reservedIDs  = [("choice",   TokenChoice),
                ("data",     TokenData),
                ("else",     TokenElse),
                ("eval",     TokenEval),
                ("external", TokenExternal),
                ("flex",     TokenFlex),
                ("free",     TokenFree),
                ("if",       TokenIf),
                ("import",   TokenImport),
                ("in",       TokenIn),
                ("infix",    TokenInfix),
                ("infixl",   TokenInfixl),
                ("infixr",   TokenInfixr),
                ("let",      TokenLet),
                ("module",   TokenModule),
                ("optmatch", TokenOptmatch),
                ("pragma",   TokenPragma),
                ("renaming", TokenRenaming),
                ("rigid",    TokenRigid),
                ("then",     TokenThen),
                ("to",       TokenTo),
                ("where",    TokenWhere)]

isIdent, isSymbol, isWhite :: Char -> Bool
isIdent  c = isAlpha c || isDigit c || c == '\'' || c == '_'
isSymbol c = elem c "~!@#$%^&*+-=<>?./|\\:"
isWhite  c = elem c " \n\r\t\v\f"

tAB_LENGTH :: Int
tAB_LENGTH = 8

-- The source location, (y,x), is the coordinates of the previous token.
-- col is the current column in the source file.  If col is 0, we are
-- somewhere at the beginning of the line before the first token.

-- Setting col to 0 is used in two places: just after emitting a virtual
-- close brace due to layout, so that next time through we check whether
-- we also need to emit a semi-colon, and at the beginning of the file,
-- to kick off the lexer.

lexer :: (Token -> P a) -> P a
lexer cont input (SrcLoc y0 x0) col | col == 0  = tab y0 x0  True  input
                                    | otherwise = tab y0 col False input -- throw away old x
  where
      -- move past whitespace and comments
      tab y x _ []            = cont EOF [] (SrcLoc y x) col
      tab y x bol ('\t'   :s) = tab y (x+ (tAB_LENGTH - (x-1) `mod` tAB_LENGTH)) bol s
      tab y _ _   ('\n'   :s) = newLine cont s y
      tab y _ _   ('-':'-':s) = newLine cont (safeTail (dropWhile (/= '\n') s)) y
         where safeTail []     = []
               safeTail (_:cs) = cs
      tab y x bol ('{':'-':s) = nestedComment tab y x bol s
      tab y x bol (c      :s)
                  | isWhite c = tab y (x+1) bol s
                  | bol       = lexBOL   cont (c:s) (SrcLoc y x) x
                  | otherwise = lexToken cont (c:s) (SrcLoc y x) x

      newLine _ s y =  tab (y+1) 1 True s

-- When we are lexing the first token of a line, check whether we need to
-- insert virtual semicolons or close braces due to layout.

lexBOL :: (Token -> P a) -> P a
lexBOL cont s loc@(SrcLoc _ x) col context
   | need_close_curly = --trace "layout: inserting '}'\n" $
                        cont TokenVRBrace s loc 0 (tail context)
                        -- Set col to 0, indicating that we're still at the
                        -- beginning of the line, in case we need a semi-colon too.
                        -- Also pop the context here, so that we don't insert
                        -- another close brace before the parser can pop it.
   | need_semi_colon = --trace "layout: inserting ';'\n" $
                       cont TokenSemicolon s loc col context
   | otherwise       = lexToken cont s loc col context
 where need_close_curly = case context of
                             []              -> False
                             (NoLayout  : _) -> False
                             (Layout n  : _) -> x < n
       need_semi_colon  = case context of
                             []              -> False
                             (NoLayout  : _) -> False
                             (Layout n  : _) -> x == n

lexToken :: (Token -> P a) -> P a
lexToken cont (c:s) (SrcLoc y x') x = --trace ("lexer: y="++show y++" x="++show x++"\n") $ 
   case c of
      -- First the special symbols
      ';'            -> special TokenSemicolon
      ','            -> special TokenComma
      '_'            -> special TokenAnonymous
      '('            -> special TokenLParen
      ')'            -> special TokenRParen
      '['            -> special TokenLBracket
      ']'            -> special TokenRBracket
      '{'            -> special TokenLBrace
      '}'            -> special TokenRBrace . tail -- pop context on '}'
      ':'            -> extend c s isSymbol     TokenInfixConstructor reservedOps
      _ | isLower  c -> extend c s isIdentOrDot maybeDottedID         reservedIDs
        | isUpper  c -> extend c s isIdentOrDot maybeDottedID         []
        | isSymbol c -> extend c s isSymbol     TokenInfixOpId        reservedOps
        | isDigit  c -> let (n,rest) = span isDigit (c:s) in forward (length n) (TokenIntLiteral n) rest
        | otherwise  -> parseError ("illegal character \'" ++ show c ++ "\'\n") s (SrcLoc y x') x
 where special t = forward 1 t s
       forward n t s' = cont t s' (SrcLoc y x') (x+n)
       extend c' s' pred def reserved = forward (1 + length tl) (fromMaybe (def cs) (lookup cs reserved)) rest
          where (tl, rest) = span pred s'
                cs = c':tl
       isIdentOrDot z = isIdent z || z == '.'
       maybeDottedID xs | '.' `elem` xs = let ys = splitAtDots xs   -- SUP: inefficient
                                          in case last ys of
                                                (z:_) | isLower z -> TokenDottedIdentifier  ys
                                                      | otherwise -> TokenDottedConstructor ys
                                                i                 -> TokenError ("Illegal identifier `" ++ i ++ "'")
                        | otherwise     = (if isLower (head xs) then TokenIdentifier else TokenConstructor) xs
       splitAtDots xs = case span (/= '.') xs of
                           (ys, '.':zs) -> ys : splitAtDots zs
                           (ys, _     ) -> [ys]

lexToken _ _ _ _ = error "Internal error: empty input in lexToken"

nestedComment :: (Int -> Int -> Bool -> String -> ParseState -> ParseResult a)
              -> (Int -> Int -> Bool -> String -> ParseState -> ParseResult a)
nestedComment cont y x bol s =
   case s of
      ('-':'}':s') -> cont                               y     (x+2) bol  s'
      ('{':'-':s') -> nestedComment (nestedComment cont) y     (x+2) bol  s'
      ('\n'   :s') -> nestedComment cont                 (y+1) 1     True s'
      (_      :s') -> nestedComment cont                 y     (x+1) bol  s'
      []           -> error "Internal error: nestedComment"
