module Machine
where

import MyPrelude
import MangleTcToCC(SearchVarTab)
import InstructionSet
import Flatten
import StateMonad
import StateComponents
import TraceClass
import Stack






interprete :: SearchVarTab -> Code -> Bool -> IO ()
---------------------------------------------------
interprete stab code traceMarker =
   do (unStateM interpreteCode) (initState code traceMarker stab)	
      return ()





interpreteCode :: StateM ()
---------------------------
interpreteCode =
   do instr <- instrAtPC
      interpreteInstr instr




interpreteInstr :: Instruction -> StateM ()
-------------------------------------------
interpreteInstr LET =
   do traceStep
      INT n    <- nextInstr
      incPCBy 2
      pointers <- allocateFrames n
      mapM_ extendEnvironment pointers
      mapM_ (\ p -> do (INT addr) <- instrAtPC
		       incPC
		       closure    <- makeClosure addr 
		       bind p closure ) pointers
      interpreteCode


interpreteInstr LETFREE =
   do traceStep
      p <- allocateFrame
      unbound p
      extendEnvironment p
      incPC
      interpreteCode


interpreteInstr ARG =
   do traceStep
      INT x <- nextInstr
      p	    <- lookupCurEnv x
      pushArgument p
      incPCBy 2
      interpreteCode

-- the following implementation of TAKE is
-- only due to the fact that we have no type
-- informations during the conpilation:
-- STRICT can enforce the interpretation
-- of expressions with functional type,
-- i.e. not saturated functions.
-- (similary, "main" may have a functional type
-- and is interpretated)
-- see comment in Compiler.hs, also

interpreteInstr TAKE =
   do traceStep
      INT x <- nextInstr
      p	    <- popArgument
      bindEnv x p
      incPCBy 2
      interpreteCode
--       ifThenElseM (argStackEmpty) (do pc <- getPC
-- 				      closure <- makeClosure pc
-- 				      setValue closure
-- 				      doRETURN)
-- 				  (do INT x <- nextInstr
-- 				      p	    <- popArgument
-- 				      bindEnv x p
-- 				      incPCBy 2
-- 				      interpreteCode)


interpreteInstr EVAL =
   do traceStep
      INT addr <- nextInstr
      closure  <- makeClosure addr
      pushReturnStack closure
      incPCBy 2
      interpreteCode

interpreteInstr BLACKHOLE_MARKER =
   do traceStep
      INT x <- nextInstr
      p	    <- lookupCurEnv x
      markEntered p
      incPCBy 2
      interpreteCode



interpreteInstr DEREF =
   do traceStep
      INT x <- nextInstr
      p	    <- lookupCurEnv x
      b	    <- lookupHeap p
      case b of
	   Unbound              -> do incPCBy 2
				      interpreteCode
	   BlackHoleMarker      -> do suspendThread_D p
				      pc <- getPC
				      actualizeThread_D pc
				      nextThread_D 
				      interpreteCode
	   Bound closure	-> do setClosure closure
				      interpreteCode

interpreteInstr RETURN = 
   do traceStep
      doRETURN


interpreteInstr VALUE =
    do traceStep
       pc      <- getPC
       closure <- makeClosure (pc+1)
       setValue closure
       doRETURN


interpreteInstr UPDATE =
   do traceStep 
      INT x <- nextInstr
      p	    <- lookupCurEnv x
      v	    <- getValue 
      bind p v
      wakeUpThreads_D p
      incPCBy 2
      interpreteCode

-- see comment for TAKE


interpreteInstr STRICT =
   do traceStep 
      Closure addr env <- getValue
      instr	       <- instrAt addr
      case instr of
	   TAKE	      -> do --setValueFun
			    incPC
			    interpreteCode
	   UVAR	      -> do incPC
			    interpreteCode
	   CONSTANT _ -> do incPC
			    interpreteCode
	   CONSTR _   -> do INT addr' <- instrAt (addr + cStrictAddrOffset)
			    setClosure (Closure addr' env)
			    interpreteCode
	   _	    -> error $ "STRICT: unexpected instruction in closure v: " ++ show instr 




interpreteInstr WAIT =
   do traceStep
      Closure addr env <- getValue
      instr <- instrAt addr
      case instr of
	   UVAR -> do INT x <- instrAt (addr + 1)
		      let p = lookupEnv env x
		      b <- lookupHeap p
		      case b of
			   Bound closure  -> do setValue closure
						interpreteCode
                           _              -> ifThenElseM runnableThreadsLeft_D
							 (do pc <- getPC
							     actualizeThread_D pc
							     suspendThread_D p
							     nextThread_D
							     interpreteCode)	
							 backtracking
	   _    -> do incPC
		      interpreteCode




interpreteInstr SWITCH  =
   do traceStep
      Closure addr env <- getValue
      INT c_id	       <- instrAt (addr + cIdOffset)
      setValue (Closure (addr + cPatternOffset) env)
      pc <- getPC
      INT alt_addr <- instrAt (pc + 1 + c_id)
      setPC alt_addr
      interpreteCode



interpreteInstr TRY =
   do traceStep
      INT alt <- nextInstr
      pushChoicePoint alt
      incPCBy 2
      interpreteCode



interpreteInstr BACKTRACK =
   do traceStep
      backtracking


interpreteInstr BOUND = 
   do traceStep
      (Closure addr env) <- getValue
      instr		 <- instrAt addr
      case instr of
	   UVAR -> do INT x <- instrAt (addr + 1)
		      let p = lookupEnv env x
		      b <- lookupHeap p
		      case b of
			   Bound closure   -> do setValue closure
						 interpreteCode 
						 -- closure could be UVAR, too
			   BlackHoleMarker -> do pc <- getPC
						 actualizeThread_D pc
						 suspendThread_D p
						 nextThread_D
						 interpreteCode
			   _		   -> do INT addr <- nextInstr
						 setPC addr
						 interpreteCode
	   _   -> do incPCBy 2
		     interpreteCode



interpreteInstr BIND =
   do traceStep
      p <- allocateFrame
      extendEnvironment p
      v@(Closure addr env) <- getValue
      bind p v
      setValue (Closure (addr + nextVarOffset) env)
      incPC
      interpreteCode



interpreteInstr BIND_TO =
   do traceStep
      Closure addr env <- getValue
      INT x <- instrAt (addr + 1)
      let p = lookupEnv env x
      INT binding <- nextInstr
      closure	  <- makeClosure binding
      bind p closure
      incPCBy 2
      interpreteCode



interpreteInstr UNIFY =
   do traceStep
      (cl1', cl2') <- childValues_D
      cl1@(Closure addr1 env1) <- hat cl1'
      cl2@(Closure addr2 env2) <- hat cl2'
      instr1 <- instrAt addr1
      instr2 <- instrAt addr2
      removeChilds_D -- in some cases (backtracking), this isn't necessary => nicer this way, but less efficient

      let occursCheck 0 _ _    _   = return False
	  occursCheck a p addr env =
	     do INT x <- instrAt (addr + 1)
		if p == lookupEnv env x then
		   return True
		 else
	           occursCheck (a-1) p (addr + nextVarOffset) env

	  createEnv 0 _    env = return env
	  createEnv a addr env =
	     do INT x <- instrAt (addr + 1)
		p <- allocateFrame
		unbound p
		let env' = bindEnv' env x p
		createEnv (a-1) (addr + nextVarOffset) env'


	  spanUThreads 1 addr_x env_x addr_y env_y =
	     do actualizeThread_D c_unify_addr 
		newThreads_D (Closure addr_x env_x) (Closure addr_y env_y)
		return ()

	  spanUThreads a addr_x env_x addr_y env_y =
	     do actualizeThread_D c_and_addr
		newNotRunnableThreads_D
		lSonThread_D
		actualizeThread_D c_unify_addr
		newThreads_D (Closure addr_x env_x) (Closure addr_y env_y)
		parentThread_D
		rSonThread_D
		spanUThreads (a-1) (addr_x + nextVarOffset) env_x (addr_y + nextVarOffset) env_y
		
      case (instr1,instr2) of

	      (UVAR, UVAR) ->
	         do INT x <- instrAt (addr1 + 1)
                    INT y <- instrAt (addr2 + 1)
		    let px = lookupEnv env1 x
                        py = lookupEnv env2 y
                    if px == py then
		       do incPC
		          interpreteCode
                     else
		       do bind px cl2
		          wakeUpThreads_D px
		          incPC
		          interpreteCode


              (UVAR, CONSTANT _) ->
	         do INT x <- instrAt (addr1 + 1)
		    let px = lookupEnv env1 x
		    bind px cl2
		    wakeUpThreads_D px
		    incPC
		    interpreteCode


	      (UVAR, CONSTR _) ->
	         do INT x     <- instrAt (addr1 + 1)
		    INT arity <- instrAt (addr2 + cArityOffset)
		    let cPatternAddr = (addr2 + cPatternOffset)
		    occurs <- occursCheck arity (lookupEnv env1 x) cPatternAddr env2
		    if occurs then
		       backtracking
		     else
		       do env' <- createEnv arity cPatternAddr env2
			  let px = lookupEnv env1 x
			  bind px cl2
			  wakeUpThreads_D px
			  spanUThreads arity cPatternAddr env2 cPatternAddr env'
			  nextThread_D
			  interpreteCode
			  


              (CONSTANT _, UVAR) ->
	         do INT x <- instrAt (addr2 + 1)
		    let px = lookupEnv env2 x
		    bind px cl1
		    wakeUpThreads_D px
		    incPC
		    interpreteCode



	      (CONSTR _, UVAR) ->
	         do INT x     <- instrAt (addr2 + 1)
		    INT arity <- instrAt (addr1 + cArityOffset)
		    let cPatternAddr = (addr1 + cPatternOffset)
		    occurs <- occursCheck arity (lookupEnv env2 x) cPatternAddr env1
		    if occurs then
		       backtracking
		     else
		       do env' <- createEnv arity cPatternAddr env1
			  let px = lookupEnv env2 x
			  bind px cl1
			  wakeUpThreads_D px
			  spanUThreads arity cPatternAddr env1 cPatternAddr env'
			  nextThread_D
			  interpreteCode

	      (CONSTANT _, CONSTANT _) ->
	         do INT id1 <- instrAt (addr1 + cIdOffset)
		    INT id2 <- instrAt (addr2 + cIdOffset)
		    if id1 == id2 then
		       do incPC
			  interpreteCode
		     else
		       backtracking



              (CONSTR _, CONSTR _) ->
	         do INT id1   <- instrAt (addr1 + cIdOffset)
		    INT id2   <- instrAt (addr2 + cIdOffset)
		    INT arity <- instrAt (addr1 + cArityOffset)
		    if id1 == id2 then
		       do spanUThreads arity (addr1 + cPatternOffset) env1 (addr2 + cPatternOffset) env2
			  nextThread_D
			  interpreteCode
		     else
		       backtracking


-- new
	      (CONSTANT _, CONSTR _) ->
	         backtracking

	      (CONSTR _, CONSTANT _) ->
	         backtracking

	      _ -> do output $ "Closures to unify:\n" ++ show cl1 ++ "\n" ++ show cl2 ++ "\n"
		      error "UNIFY: unrecognized pattern in closure!\n"
			   


interpreteInstr A_THREADS =
   do traceStep
      INT t2_addr <- nextInstr
      pc <- getPC
      c1 <- makeClosure (pc+2)
      c2 <- makeClosure t2_addr
      actualizeThread_D c_and_addr
      newThreads_D c1 c2
      nextThread_D
      interpreteCode


interpreteInstr U_THREADS =
   do traceStep
      INT t2_addr <- nextInstr
      pc <- getPC
      c1 <- makeClosure (pc+2)
      c2 <- makeClosure t2_addr
      actualizeThread_D c_unify_addr
      newThreads_D c1 c2
      nextThread_D
      interpreteCode



interpreteInstr WAKE_UP_THREADS =
   do traceStep
      Closure addr env <- getValue
      INT x <- instrAt (addr+1)
      let p = lookupEnv env x
      wakeUpThreads_D p
      incPC
      interpreteCode


interpreteInstr TERM_THREADS =
   do traceStep
      removeChilds_D
      incPC
      interpreteCode


interpreteInstr AWAIT_SONS =
   do traceStep
      b <- childsTerminated_D
      if b then
	 do incPC
	    interpreteCode
       else
	 do firstChildTerminated_D
	    nextThread_D -- no actualization necessary, as code will be reached never again TODO think of it again
	    interpreteCode

interpreteInstr UVAR =
   do traceStep
      pc      <- getPC
      closure <- makeClosure pc
      setValue closure
      doRETURN


interpreteInstr (CONSTR _) =
   do traceStep
      pc      <- getPC
      closure <- makeClosure pc
      setValue closure
      doRETURN


interpreteInstr (CONSTANT _) =
   do traceStep
      pc      <- getPC
      closure <- makeClosure pc
      setValue closure
      doRETURN


interpreteInstr (ERROR e) =
   error e

interpreteInstr code =
   error $ "interpreteInstr: no interpretation possible for: " ++ show code ++ "\n"



doRETURN :: StateM ()
---------------------
doRETURN =
   do b <- returnStackEmpty
      if b then
	 do t <- parentThreadExists_D
	    if t then
	       do actualizeThreadsValue_D
		  parentThread_D
		  interpreteCode
	     else
	       do more <- answer
		  if more then
		     backtracking
		   else
		     do traceInstrCount
			return ()
       else
	 do closure <- popReturnStack
	    setClosure closure
	    interpreteCode


backtracking :: StateM ()
-------------------------
backtracking =
   do poss <- furtherChoicepoints
      if poss then
	 do backtrack
	    interpreteCode
       else
         do output ("no (more) solutions!\n")
	    traceInstrCount
	    return ()

