module MangleRdrToTc(rdrModuleToTcModule) where

import Monad(foldM, unless, liftM, liftM2)
import List(groupBy, group, sort)
import FiniteMap(FiniteMap, emptyFM, addToFM, addListToFM, lookupFM, lookupWithDefaultFM)
import Pretty((<+>), text, render)
import CommonSyntax(Pretty, sepByCommas, MatchStrategy, Annotation)
import ReaderSyntax
import PrettyReaderSyntax
import TypeCheckerSyntax
import PrettyTypeCheckerSyntax
import ErrorMonad

rdrModuleToTcModule :: MatchStrategy -> Annotation -> RdrModule -> MaybeError TcModule
rdrModuleToTcModule defMStrat                  -- default matching strategy
                    defAnnot                   -- default annotation
                    (RdrModule  _              -- ignoring module name and exports
                               (_,             -- ignoring imports
                                pragmas,
                                _,             -- ignoring fixity declarations
                                blockDecls)) = do
   (mStrat, annot) <- collectPragmas pragmas defMStrat defAnnot
   let dds  = [ mangleDataDeclaration dd | RdrDataDecl dd <- blockDecls ]
   mfds <- mangleFunctionDeclarations annot [ fd | RdrFuncDecl fd <- blockDecls ]
   return (TcModule mStrat dds mfds)

----------------------------------------------------------------------
-- Calculate matching strategy and default annotation for this module.

collectPragmas :: [RdrPragma] -> MatchStrategy -> Annotation -> MaybeError (MatchStrategy,Annotation)
collectPragmas pragmas defMStrat defAnnot = do
   mStrat <- checkPragmas "matching strategy"     [ m | RdrPragmaMatch m <- pragmas ] defMStrat
   annot  <- checkPragmas "evaluation annotation" [ a | RdrPragmaAnnot a <- pragmas ] defAnnot
   return (mStrat, annot)

checkPragmas :: Pretty a => String -> [a] -> a -> MaybeError a
checkPragmas _   []         def = return def
checkPragmas _   [p]        _   = return p
checkPragmas msg ps@(_:_:_) _   = wrong ("Multiple " ++ msg ++ " pragmas:") ps

----------------------------------------------------------------------
-- Merge equations, signatures, and annotations together. We go to some pains
-- here to keep the order of function declarations. This is not neccessary for
-- the semantics, but makes things more readable for debugging.

-- Merging is done via a map from function names to their declaration.
type DeclMap = FiniteMap RdrName TcLocalDef

mangleFunctionDeclarations :: Annotation -> [RdrFunctionDeclaration] -> MaybeError [TcLocalDef]
mangleFunctionDeclarations defAnnot fds = do
   let signatures           = [ (name,typ) | RdrSignature names typ <- fds, name <- names ]
       annotations          = [ (name,ann) | RdrEvalAnnot name ann <- fds ]
       equations            = [ (name,eq)  | RdrEquation  eq@(RdrEquat name _ _ _) <- fds ]
       equationGroups       = groupTogether equations
       funcNames            = map fst equationGroups
       makeFDef (name, eqs) = do
          meqs <- mapM (mangleEquat defAnnot) eqs
          return (name, TcLocalFuncDecl Nothing Nothing (mangleName name) meqs)
   checkDuplicates "Multiple definitions for" funcNames
   mequationGroups <- mapM makeFDef equationGroups
   let fm1 = addListToFM emptyFM mequationGroups
   fm2 <- foldM mergeType  fm1 signatures
   fm3 <- foldM mergeAnnot fm2 annotations
   let mfds = map (mergeDefaultAnnotation defAnnot .
                   lookupWithDefaultFM fm3 (error "impossible")) funcNames
   mapM_ checkArity mfds
   mapM_ checkLinearity mfds
   return mfds

mergeType :: DeclMap -> (RdrName,RdrType) -> MaybeError DeclMap
mergeType fm (n,t) =
   case lookupFM fm n of
      Nothing ->
         wrong "Type signature but not equation for" [n]
      Just (TcLocalFuncDecl (Just _) _ _ _) ->
         wrong "Multiple type signatures for" [n]
      Just (TcLocalFuncDecl Nothing a name eqs) ->
         return (addToFM fm n (TcLocalFuncDecl (Just (mangleType t)) a name eqs))
      Just (TcLocalPatternDecl _ _) ->
         error "impossible (mergeType)"

mergeAnnot :: DeclMap -> (RdrName,Annotation) -> MaybeError DeclMap
mergeAnnot fm (n,a) =
   case lookupFM fm n of
      Nothing ->
         wrong "Evaluation annotation but not equation for" [n]
      Just (TcLocalFuncDecl _ (Just _) _ _) ->
         wrong "Multiple evaluation annotations for" [n]
      Just (TcLocalFuncDecl t Nothing name eqs) ->
         return (addToFM fm n (TcLocalFuncDecl t (Just a) name eqs))
      Just (TcLocalPatternDecl _ _) ->
         error "impossible (mergeAnnot)"

mergeDefaultAnnotation :: Annotation -> TcLocalDef -> TcLocalDef
mergeDefaultAnnotation defAnnot (TcLocalFuncDecl mbType Nothing name equations) =
   TcLocalFuncDecl mbType (Just defAnnot) name equations
mergeDefaultAnnotation _ localDef = localDef

----------------------------------------------------------------------
-- Check equations of a function for consistent arity.

checkArity :: TcLocalDef -> MaybeError ()
checkArity (TcLocalFuncDecl _ _ name eqs) =
   unless ((length . group . map getFuncArity $ eqs) == 1)
          (wrong "Varying number of arguments for function" [name])
   where getFuncArity (TcEquat pats _) = length pats
checkArity (TcLocalPatternDecl _ _) = return ()

----------------------------------------------------------------------
-- Check patterns for linearity.

checkLinearity :: TcLocalDef -> MaybeError ()
checkLinearity (TcLocalFuncDecl _ _ (TcName name) eqs) = mapM_ checkLin eqs
   where checkLin (TcEquat patterns _) =
            checkDuplicates
               ("Pattern for " ++ name ++ " is not linear, repeated variable(s):")
               (concatMap collectVars patterns)
         collectVars (TcPatternVar v)       = [v]
         collectVars TcPatternAnonymous     = []
         collectVars (TcPatternLiteral _)   = []
         collectVars (TcPatternAppl _ pats) = concatMap collectVars pats
checkLinearity p@(TcLocalPatternDecl _ _) =
   wrong "pattern declarations not yet implemented:" [p]

----------------------------------------------------------------------

mangleDataDeclaration :: RdrDataDeclaration -> TcDataDeclaration
mangleDataDeclaration (RdrDataDeclaration c vs cs) =
   TcDataDeclaration (mangleName c) (map mangleName vs) (map mangleConstrDecl cs)

mangleConstrDecl :: RdrConstrDecl -> TcConstrDecl
mangleConstrDecl (RdrConstrDecl c ts) =
   TcConstrDecl (mangleName c) (map mangleType ts)

mangleType :: RdrType -> TcType
mangleType (RdrTypeVar    v)   = TcTypeVar (mangleName v)
mangleType (RdrTypeConstr c)   = TcTypeConstr (mangleQName c) []
mangleType (RdrTypeFunc   f a) = TcTypeConstr funcName [mangleType f, mangleType a]
mangleType t@(RdrTypeAppl _ _) = foo t []
   where foo (RdrTypeAppl f a) accu = foo f (mangleType a : accu)
         foo (RdrTypeConstr c) accu = TcTypeConstr (mangleQName c) accu
         foo _                 _    = error "impossible (mangleType)"
mangleType (RdrTypeTuple  ts)  = TcTypeConstr (tupleName ts) (map mangleType ts)
mangleType (RdrTypeList   t)   = TcTypeConstr listName [mangleType t]

mangleEquat :: Annotation -> RdrEquat -> MaybeError TcEquat
mangleEquat defAnnot (RdrEquat _ pats rhs whereDecls) =
   liftM (TcEquat (map manglePattern pats))
         (mangleLocalDefs defAnnot whereDecls =<< mangleFunRHS defAnnot rhs)

mangleFunRHS :: Annotation -> RdrFunRHS -> MaybeError TcExpr
mangleFunRHS defAnnot (RdrUnguardedRHS rhs)  = mangleExpr defAnnot rhs
mangleFunRHS defAnnot (RdrGuardedRHSs gdrhs) =
   liftM (TcExprAppl (TcExprVar guardName))
         (mangleExpr defAnnot (RdrExprList [ RdrExprTuple [g, e] | RdrGuardedRHS g e <- gdrhs ]))

manglePattern :: RdrPattern -> TcPattern
manglePattern (RdrPatternVar v) =
  TcPatternVar (mangleName v)
manglePattern RdrPatternAnonymous =
   TcPatternAnonymous
manglePattern (RdrPatternLiteral l) =
   TcPatternLiteral l
manglePattern (RdrPatternAppl (RdrUnqualName ":") pats) = -- SUP: HACK HACK HACK
   TcPatternAppl consName (map manglePattern pats)
manglePattern (RdrPatternAppl c pats) =
   TcPatternAppl (mangleQName c) (map manglePattern pats)
manglePattern (RdrPatternTuple pats) =
   TcPatternAppl (tupleName pats) (map manglePattern pats)
manglePattern (RdrPatternList pats) =
   foldr (\e p -> TcPatternAppl consName [e, p])
         (TcPatternAppl nilName [])
         (map manglePattern pats)

mangleLocalDefs :: Annotation -> [RdrLocalDef] -> TcExpr -> MaybeError TcExpr
mangleLocalDefs defAnnot localDefs expr = do
   let funcDecls = [ fd  | RdrLocalFuncDecl fd            <- localDefs ]
       patDecls  = [ pd  | pd@(RdrLocalPatternDecl _ _ _) <- localDefs ]
       freeVars  = [ var | RdrLocalFreeDecl vars          <- localDefs, var <- vars ]
   unless (null patDecls)
          (wrong "pattern declarations not yet implemented:" patDecls)
   mfuncDecls <- mangleFunctionDeclarations defAnnot funcDecls
   let mkFreeDecl []  e = e
       mkFreeDecl fvs e = TcExprFreeDecl fvs e

       mkLetrec [] e = e
       mkLetrec ls e = TcExprLetrec ls e
   return (mkFreeDecl (map mangleName freeVars) (mkLetrec mfuncDecls expr))

mangleExpr :: Annotation -> RdrExpr -> MaybeError TcExpr
mangleExpr defAnnot (RdrExprLambda  patterns body) =
   liftM (TcExprLambda (map manglePattern patterns)) (mangleExpr defAnnot body)
mangleExpr defAnnot (RdrExprLet localDefs expr) =
   mangleLocalDefs defAnnot localDefs =<< mangleExpr defAnnot expr
mangleExpr defAnnot (RdrExprIf i t e) =
   liftM (foldl TcExprAppl (TcExprVar ifName)) (mapM (mangleExpr defAnnot) [i,t,e])
mangleExpr defAnnot (RdrExprAppl (RdrExprVar (RdrUnqualName ":")) arg) = -- SUP: HACK HACK HACK
   liftM (TcExprAppl (TcExprConstr consName)) (mangleExpr defAnnot arg)
mangleExpr defAnnot (RdrExprAppl func arg) =
   liftM2 TcExprAppl (mangleExpr defAnnot func) (mangleExpr defAnnot arg)
mangleExpr _ (RdrExprVar v) =
   return (TcExprVar (mangleQName v))
mangleExpr _ (RdrExprConstr c) =
   return (TcExprConstr (mangleQName c))
mangleExpr defAnnot (RdrExprTuple exprs) =
   liftM (foldl TcExprAppl (TcExprConstr (tupleName exprs))) (mapM (mangleExpr defAnnot) exprs)
mangleExpr defAnnot (RdrExprList exprs) =
   liftM (foldr (\e -> TcExprAppl (TcExprAppl (TcExprConstr consName) e))
                (TcExprConstr nilName))
         (mapM (mangleExpr defAnnot) exprs)
mangleExpr _ (RdrExprLiteral l) = return (TcExprLiteral l)

mangleQName :: RdrQName -> TcQName
mangleQName (RdrQualName m s) = TcQualName (mangleModuleID m) s
mangleQName (RdrUnqualName s) = TcUnqualName s

mangleModuleID :: RdrModuleID -> TcModuleID
mangleModuleID (RdrModuleID ns) = TcModuleID ns

mangleName :: RdrName -> TcName
mangleName (RdrName n) = TcName n

----------------------------------------------------------------------
-- special identifiers

listName :: TcQName
listName = TcUnqualName "_List"

funcName :: TcQName
funcName = TcUnqualName "_->"

tupleName :: [a] -> TcQName
tupleName components = TcUnqualName ("_Tuple" ++ show (length components))

nilName :: TcQName
nilName = TcUnqualName "_Nil"

consName :: TcQName
consName = TcUnqualName "_Cons"

ifName :: TcQName
ifName = TcUnqualName "_if"

guardName :: TcQName
guardName = TcUnqualName "_guard"

----------------------------------------------------------------------
-- error utilities

checkDuplicates :: (Ord a, Pretty a) => String -> [a] -> MaybeError ()
checkDuplicates msg xs = case dups xs of
                            [] -> return ()
                            ds -> wrong msg ds

wrong :: Pretty a => String -> [a] -> MaybeError b
wrong msg xs = fail (render (text msg <+> sepByCommas undefined xs))

----------------------------------------------------------------------
-- list utilities

-- something similar to groupBy for association lists,
-- e.g., groupTogether [('a',1),('a',2),('b',3),('a',4)] == [('a',[1,2]),('b',[3]),('a',[4])]
groupTogether :: Eq k => [(k,v)] -> [(k,[v])]
groupTogether = map (\xs -> (fst (head xs), map snd xs)) . groupBy (\(x,_) (y,_) -> x == y)

dups :: Ord a => [a] -> [a]
dups xs = [ x | (x:_:_) <- group (sort xs) ]
