module PrettyReaderSyntax where

import List(intersperse)
import Pretty
import CommonSyntax
import ReaderSyntax

---------- module structure ----------------------------------------------------------------------

instance Pretty RdrModule where
   pretty s m = case m of
                   RdrModule Nothing          ipfb -> prettyIPFB ipfb
                   RdrModule (Just (n,mbExp)) ipfb -> text "module" <+> pretty s n <>
                                                      maybe empty (parenCommas s) mbExp $$
                                                      prettyIPFB ipfb
      where prettyIPFB (i,p,f,b) = vcat [vcatP s i, vcatP s p, vcatP s f, vcatP s b]

instance Pretty RdrExport where
   pretty s (RdrExportFunc f)         = pretty s f
   pretty s (RdrExportTypeAbstract t) = pretty s t
   pretty s (RdrExportTypeAll t)      = pretty s t <> text "(..)"
   pretty s (RdrExportTypeWith t cs)  = pretty s t <> parenCommas s cs
   pretty s (RdrExportModule m mbIEs) = pretty s m <> maybe empty (parenCommas s) mbIEs

instance Pretty RdrImportEntity where
   pretty s (RdrImportFunc f) = pretty s f
   pretty s (RdrImportType t) = pretty s t

instance Pretty RdrImport where
   pretty s (RdrImport m mbIEs rs) = text "import" <+> pretty s m <+>
                                     maybe empty (parenCommas s) mbIEs <+> sepP s rs

instance Pretty RdrRenaming where
   pretty s (RdrRenaming f t) = text "renaming" <+> pretty s f <+> text "to" <+> pretty s t

instance Pretty RdrPragma where
   pretty s (RdrPragmaAnnot a) = text "pragma" <+> pretty s a
   pretty s (RdrPragmaMatch m) = text "pragma" <+> pretty s m

instance Pretty RdrBlockDecl where
   pretty s (RdrDataDecl d) = pretty s d
   pretty s (RdrFuncDecl f) = pretty s f

---------- declaration of a new algebraic type ---------------------------------------------------

instance Pretty RdrDataDeclaration where
   pretty s (RdrDataDeclaration c vs cs) =
      text "data" <+> pretty s c <+> sepP s vs <+> equals <+>
      (hcat . intersperse (text " | ") . map (pretty s)) cs

instance Pretty RdrConstrDecl where
   pretty s (RdrConstrDecl c ts) = pretty s c <+> sepP s ts

---------- type ----------------------------------------------------------------------------------

instance Pretty RdrType where
   pretty s (RdrTypeVar v)      = pretty s v
   pretty s (RdrTypeConstr c)   = pretty s c
   pretty s t@(RdrTypeFunc _ _) = parens (foo t) -- SUP: visual hack
      where foo (RdrTypeFunc f a) = pretty s f <+> text "->" <+> foo a
            foo x                 = pretty s x
   pretty s t@(RdrTypeAppl _ _) = parens (foo t) -- SUP: visual hack
      where foo (RdrTypeAppl f a) = foo f <+> pretty s a
            foo x                 = pretty s x
   pretty s (RdrTypeTuple ts)   = parenCommas s ts
   pretty s (RdrTypeList t)     = brackets (pretty s t) 

---------- fixity declaration --------------------------------------------------------------------

instance Pretty RdrFixityDecl where
   pretty s (RdrFixityDecl f ops) = pretty s f <+> sepByCommas s ops

---------- function declaration et al ------------------------------------------------------------

instance Pretty RdrFunctionDeclaration where
   pretty s (RdrSignature ns t) = sepByCommas s ns <+> text "::"   <+> pretty s t
   pretty s (RdrEvalAnnot n a)  = pretty s n <+> text "eval" <+> pretty s a
   pretty s (RdrEquation e)     = pretty s e

---------- equation ------------------------------------------------------------------------------

instance Pretty RdrEquat where
   pretty s (RdrEquat n ps r ls) =
      pretty s n <+> sepP s ps <+> pretty s r $$ prettyLocalDefs s ls

instance Pretty RdrFunRHS where
   pretty s (RdrUnguardedRHS e) = equals <+> pretty s e
   pretty s (RdrGuardedRHSs gs) = vcatP s gs

instance Pretty RdrGuardedRHS where
   pretty s (RdrGuardedRHS g e) = text "|" <+> pretty s g <+> equals <+> pretty s e

instance Pretty RdrPattern where
   pretty s (RdrPatternVar v)     = pretty s v
   pretty _ RdrPatternAnonymous   = text "_"
   pretty s (RdrPatternLiteral l) = pretty s l
   pretty s (RdrPatternAppl c []) = pretty s c   -- SUP: special case, only for nice layout
   pretty s (RdrPatternAppl c [x,y])             -- SUP: dito
                   | isOperator c = parens (pretty s x <+> pretty s c <+> pretty s y)
   pretty s (RdrPatternAppl c ps) = parens (pretty s c <+> sepP s ps)
   pretty s (RdrPatternTuple ps)  = parenCommas s ps
   pretty s (RdrPatternList ps)   = brackets . sepByCommas s $ ps

instance Pretty RdrLocalDef where
   pretty s (RdrLocalFuncDecl f)         = pretty s f
   pretty s (RdrLocalPatternDecl p e ls) = pretty s p <+> equals <+> pretty s e $$ prettyLocalDefs s ls
   pretty s (RdrLocalFreeDecl vs)        = sepP s vs <+> text "free"

prettyLocalDefs :: a -> [RdrLocalDef] -> Doc
prettyLocalDefs _ [] = empty
prettyLocalDefs s ls = nest 3 (text "where" $$ nest 6 (vcatP s ls))

---------- expressions ---------------------------------------------------------------------------

instance Pretty RdrExpr where
   pretty s (RdrExprLambda ps e) = parens (text "\\" <> sepP s ps <+> text "->" <+> pretty s e)
   pretty s (RdrExprLet ls e)    = parens (text "let" $$ nest 4 (vcatP s ls) $$
                                           text "in" <+> pretty s e)
   pretty s (RdrExprIf i t e)    = parens (sep [text "if",   pretty s i,
                                                text "then", pretty s t,
                                                text "else", pretty s e])
   pretty s (RdrExprAppl (RdrExprAppl (RdrExprVar f) a) b) -- SUP: special case, only for nice layout
                  | isOperator f = parens (pretty s a <+> pretty s f <+> pretty s b)
   pretty s e@(RdrExprAppl _ _)  = parens (foo e)
      where foo (RdrExprAppl f a) = foo f <+> pretty s a
            foo x                 = pretty s x
   pretty s (RdrExprVar v)
                  | isOperator v = parens (pretty s v)
                  | otherwise    = pretty s v
   pretty s (RdrExprConstr c)    = pretty s c
   pretty s (RdrExprTuple es)    = parenCommas s es
   pretty s (RdrExprList es)     = brackets . sepByCommas s $ es
   pretty s (RdrExprLiteral l)   = pretty s l

---------- names ---------------------------------------------------------------------------------

instance Pretty RdrQName where
   pretty s (RdrQualName m n) = pretty s m <> text ('.':n)
   pretty _ (RdrUnqualName n) = text n

instance Pretty RdrName where
   pretty _ (RdrName n) = text n

instance Pretty RdrModuleID where
   pretty _ (RdrModuleID ss) = text . concat . intersperse "." $ ss
