module ReaderSyntax where

import Char(isAlpha)
import CommonSyntax

---------- module structure ----------------------------------------------------------------------

data RdrModule =
     RdrModule (Maybe (RdrModuleID,
                       Maybe [RdrExport]))                         -- Nothing = export everything
               ([RdrImport],
                [RdrPragma],
                [RdrFixityDecl],
                [RdrBlockDecl])

data RdrExport =
     RdrExportFunc         RdrFunctionName                         -- function
   | RdrExportTypeAbstract RdrTypeConstrID                         -- T
   | RdrExportTypeAll      RdrTypeConstrID                         -- T(..)
   | RdrExportTypeWith     RdrTypeConstrID [RdrDataConstrID]       -- T(C_1,...,C_n)
   | RdrExportModule       RdrModuleID (Maybe [RdrImportEntity])   -- module (Nothing = re-export everything)

data RdrImportEntity =
     RdrImportFunc RdrFunctionName
   | RdrImportType RdrTypeConstrID

data RdrImport =
     RdrImport RdrModuleID
               (Maybe [RdrImportEntity])                           -- Nothing = import everything
               [RdrRenaming]

data RdrRenaming =
     RdrRenaming RdrName RdrName

data RdrPragma =
     RdrPragmaAnnot Annotation
   | RdrPragmaMatch MatchStrategy

data RdrBlockDecl =
     RdrDataDecl RdrDataDeclaration
   | RdrFuncDecl RdrFunctionDeclaration

---------- declaration of a new algebraic type ---------------------------------------------------

data RdrDataDeclaration =
     RdrDataDeclaration RdrTypeConstrID [RdrTypeVarID] [RdrConstrDecl]

data RdrConstrDecl =
     RdrConstrDecl RdrDataConstrID [RdrType]

---------- type ----------------------------------------------------------------------------------

data RdrType =
     RdrTypeVar    RdrTypeVarID
   | RdrTypeConstr RdrQTypeConstrID
   | RdrTypeFunc   RdrType RdrType
   | RdrTypeAppl   RdrType RdrType
   | RdrTypeTuple  [RdrType]
   | RdrTypeList   RdrType

---------- fixity declaration --------------------------------------------------------------------

data RdrFixityDecl =
     RdrFixityDecl Fixity [RdrInfixOpID]

---------- function declaration et al ------------------------------------------------------------

data RdrFunctionDeclaration =
     RdrSignature [RdrFunctionName] RdrType
   | RdrEvalAnnot RdrFunctionName Annotation
   | RdrEquation  RdrEquat

---------- equation ------------------------------------------------------------------------------

data RdrEquat =
     RdrEquat RdrFunctionName [RdrPattern] RdrFunRHS [RdrLocalDef]

data RdrFunRHS =
     RdrUnguardedRHS RdrExpr
   | RdrGuardedRHSs  [RdrGuardedRHS]

data RdrGuardedRHS =
     RdrGuardedRHS RdrExpr RdrExpr

data RdrPattern =
     RdrPatternVar       RdrVariableID
   | RdrPatternAnonymous
   | RdrPatternLiteral   Literal
   | RdrPatternAppl      RdrQDataConstrID [RdrPattern]
   | RdrPatternTuple     [RdrPattern]
   | RdrPatternList      [RdrPattern]

data RdrLocalDef =
     RdrLocalFuncDecl    RdrFunctionDeclaration
   | RdrLocalPatternDecl RdrPattern RdrExpr [RdrLocalDef]
   | RdrLocalFreeDecl    [RdrVariableID]

---------- expressions ---------------------------------------------------------------------------

data RdrExpr =
     RdrExprLambda  [RdrPattern] RdrExpr
   | RdrExprLet     [RdrLocalDef] RdrExpr
   | RdrExprIf      RdrExpr RdrExpr RdrExpr
   | RdrExprAppl    RdrExpr RdrExpr
   | RdrExprVar     RdrQVariableID   -- NOTE: Could be a function, too
   | RdrExprConstr  RdrQDataConstrID
   | RdrExprTuple   [RdrExpr]
   | RdrExprList    [RdrExpr]
   | RdrExprLiteral Literal

---------- names ---------------------------------------------------------------------------------

data RdrQName =
     RdrQualName   RdrModuleID String
   | RdrUnqualName String

isOperator :: RdrQName -> Bool
isOperator (RdrQualName _ (c:_)) = not (isAlpha c)
isOperator (RdrUnqualName (c:_)) = not (isAlpha c)
isOperator _                     = True

newtype RdrName = RdrName String    deriving (Eq,Ord)

newtype RdrModuleID = RdrModuleID [String]

type RdrQVariableID   = RdrQName
type RdrVariableID    = RdrName

type RdrQDataConstrID = RdrQName
type RdrDataConstrID  = RdrName

type RdrQTypeConstrID = RdrQName
type RdrTypeConstrID  = RdrName

type RdrFunctionName  = RdrName
type RdrInfixOpID     = RdrName

type RdrTypeVarID     = RdrName
