module StateComponents
where

import MyPrelude

import MangleTcToCC(SearchVarTab)
import InstructionSet
import Flatten(c_start_addr)
import Stack
import Queue
import FiniteMap
import TraceClass


data State = State {
		     code	 :: Code
		   , pc		 :: CAddr
		   , heap        :: Heap
		   , env         :: Environment
		   , dispatcher  :: Dispatcher
		   , chpS	 :: ChoicepointStack
		   , v		 :: Value
		   , a		 :: ArgStack
		   , r		 :: ReturnStack
		   ----------------------------
		   , instrCount	 :: Int                    -- instruction counter
		   , traceNum	 :: Int
		   , traceMarker :: Bool		   -- should the execution be traced?
		   , stab	 :: SearchVarTab	   -- the search variables (necessary for output)
		   }





initState code' traceMarker' stab'
	 = State {
		    code        = code'
		  , pc		= c_start_addr
		  , heap        = emptyHeap
		  , env		= emptyEnv
		  , dispatcher	= emptyDispatcher
		  , chpS	= emptyStack
		  , v		= emptyClosure
		  , a		= emptyStack
		  , r		= emptyStack
		  ---------------------------
		  , instrCount  = 0
		  , traceNum	= 0
		  , traceMarker = traceMarker'
		  , stab	= stab'
		  }





instance Trace State where
   trace state =
      do putStrLn "1.    Inspect code"
	 putStrLn "2.    Inspect heap"
	 putStrLn "3.    Inspect environment"
	 putStrLn "4.    Inspect dispatcher"
	 putStrLn "5.    Inspect choice point stack"
	 putStrLn "6.    Inspect v"
	 putStrLn "7.    Inspect a"
	 putStrLn "8.    Inspect r"
	 putStrLn "Else: proceed"
	 putStr   "-> "
	 answer <- getInt
	 traceNl
	 case answer of
	    1  -> do trace (code state)
		     traceNl
		     trace state
	    2  -> do trace (heap state)
		     traceNl
		     trace state
	    3  -> do trace (env state)
		     traceNl
		     trace state
	    4  -> do trace (dispatcher state)
		     traceNl
		     trace state
	    5  -> do trace (chpS state)
		     traceNl
		     trace state
	    6  -> do trace (v state)
		     traceNl
		     trace state
	    7  -> do trace (a state)
		     traceNl
		     trace state
	    8  -> do trace (r state)
		     traceNl
		     trace state
	    _  ->    return ()


-- Heap --
----------

data HeapAddr    = HeapAddr Int
     deriving (Ord,Eq,Show)

data Heap        = Heap [HeapAddr] (FiniteMap HeapAddr HeapBinding)
     
instance Show Heap where
     showsPrec _ (Heap _ map) = showString "Heap:\t" . shows (fmToList map)

data HeapBinding  = 
     Bound Closure | 
     Unbound | 
     BlackHoleMarker
     deriving (Show)


emptyHeap = Heap (map HeapAddr [0..]) emptyFM


allocateFrame' (Heap (addr:addrs) map) = 
   (Heap addrs map, addr)
allocateFrame' (Heap [] _) = error "StateComponents.allocateFrame': Impossible error!\n"

allocateFrames' (Heap addrs map) n = 
   (Heap (drop n addrs) map, take n addrs)


unbound' (Heap addrs map) p =
   (Heap addrs (addToFM map p Unbound))


bind' (Heap addrs map) p c =
   (Heap addrs (addToFM map p (Bound c)))


markEntered' (Heap addrs map) p =
   (Heap addrs (addToFM map p BlackHoleMarker))


lookupHeap'  (Heap _ map) p =
   lookupWithDefaultFM map
		       (error "Heap: Lookup for uninitialized address!\n")
		       p


hat' code heap closure@(Closure cAddr env) =
   let instr   = lookupCode code cAddr
   in  if instr == UVAR then
	  let (INT v) = lookupCode code (cAddr+1)
	      binding = lookupHeap' heap (lookupEnv env v)
	  in  case binding of
		   Bound closure'  -> hat' code heap closure'
		   Unbound	   -> closure
		   BlackHoleMarker -> error "hat: blackhole marker!\n"
        else closure





instance Trace HeapAddr where
   trace (HeapAddr a) = putStr $ " -> " ++ show a

instance Trace Heap where
   trace (Heap addrs map) =
	        do let (HeapAddr a) = head addrs
		   putStrLn $ "Currently allocated frames: " ++ show a
		   putStrLn "Which heap address do you want to inspect?"
		   putStr "-> "
		   p <- getInt
		   case lookupFM map (HeapAddr p) of
			Nothing -> putStr "Inaccessible address."
			Just b  -> do putStrLn $ "Content of address " ++ show p ++ ":"
				      case b of
					   Unbound         -> putStr "Nil"
					   BlackHoleMarker -> putStr "Blackhole marker"
					   Bound closure   -> trace closure



-- Environment --
-----------------


data Environment = Environment [HeapAddr]
     deriving (Show)

emptyEnv = Environment []

extendEnvironment' (Environment e) p =
   Environment $ e ++ [p]

lookupEnv (Environment e) x =
   if x >= (length e) then
      error $ "lookupEnv: index (" ++ show x ++ ") out of range:\n" ++ show e ++ "\n"
    else
      e!!x

bindEnv' (Environment e) x p =
   if x > (length e) then
      error $ "bindEnv': index (" ++ show x ++ ") out of range:\n" ++ show e ++ "\n"
    else
      let pref = take x e
	  post = drop (x+1) e
      in  Environment $ pref ++ [p] ++ post

instance Trace Environment where
   trace (Environment e) = do putStrLn $ "Environment: " 
			      let tEnv :: (Int, HeapAddr) -> IO()
                                  tEnv (n,a) = do putChar '\t'
						  putStr $ show n
						  trace a
						  traceNl
			      mapM_ tEnv (zip [0..] e)




-- Dispatcher --
----------------

-- TODO modul abhaengigkeiten

data Thread = Thread Closure Value ArgStack ReturnStack

type ThreadId = Int

type Runnable   = Queue(ThreadId)
type Suspended  = FiniteMap HeapAddr [ThreadId]
type ThreadTree = FiniteMap ThreadId (ThreadId, ThreadId, ThreadId)
type Content    = FiniteMap ThreadId Thread
type SonT	= [ThreadId]
type Delta	= ThreadId
type FreeIds	= [ThreadId]

data Dispatcher = Dispatcher FreeIds Runnable Suspended ThreadTree Content SonT Delta

--                             f        r        s                t                  c    st  d
--                             |        |        |                |                  |     |  |
emptyDispatcher = Dispatcher [1..] emptyQueue emptyFM (addToFM emptyFM 0 (0,0,-1)) emptyFM [] 0


instance Trace Dispatcher where
   trace (Dispatcher _ r s _ c st d) =
      do putStr $ "Current thread id:\t\t\t\t\t"
	 trace d
	 traceNl
	 putStr   $ "First child of the following threads terminated:\t"
	 traceSepSpace st
	 traceNl
	 putStr   $ "Runnable threads:\t\t\t\t\t"
	 trace r
	 traceNl
	 putStr   $ "Suspended threads:\t\t\t\t\t"
	 traceSepSpace (concat $ eltsFM s)
	 traceNl
	 putStr   $ "All instantiated threads:\t\t\t\t"
	 traceSepSpace (keysFM c)
	 

instance Trace Int where
   trace i = putStr $ show i


		
-- Closure --
-------------

data Closure = 
     Closure CAddr Environment
     deriving (Show)     

emptyClosure = Closure 0 emptyEnv

instance Trace Closure where
   trace (Closure pc e) = do putStr "Closure: code address: "
			     trace pc
			     traceNl
			     trace e  

type Value       = Closure

type ArgStack    = Stack HeapAddr

type ReturnStack = Stack Closure




data Choicepoint = Choicepoint CAddr Heap Environment Dispatcher Value ArgStack ReturnStack

type ChoicepointStack = Stack Choicepoint


pushChoicePoint' :: State -> CAddr -> ChoicepointStack
------------------------------------------------------
pushChoicePoint' s pc = push (chpS s) (Choicepoint pc (heap s) (env s) (dispatcher s) (v s) (a s) (r s))

backtrack' :: State -> State
----------------------------
backtrack' s
  | isEmptyStack cs = error "backtrack': choice point stack empty!\n"
  | otherwise	    = let (Choicepoint pc' h e d v' a' r') = top cs
			  chpS'				   = pop cs
		      in         s{ pc	       = pc'
				  , heap       = h
			 	  , env        = e
				  , dispatcher = d
				  , chpS       = chpS'
				  , v	       = v'
				  , a	       = a'
				  , r	       = r' }
  where cs = (chpS s)


instance Trace Choicepoint where
   trace _ = putStr "Sorry: trace choicepoint not yet implemented!\n"
-- TODO