module StateMonad
where

import MyPrelude

import InstructionSet
import Flatten(  cArityOffset
	       , cIdOffset
	       , cPatternOffset
	       )
import Stack
import StateComponents
import DispatcherPro
import TraceClass
import List
import Maybe



newtype StateM a = StateM (State -> IO (a, State))



instance Monad StateM where
   m >>= k    = StateM (\ state -> 
			   do (val, state') <- (unStateM m) state
			      unStateM (k val) state')
   return val = StateM (\ state -> return (val, state))

unStateM (StateM m) = m


traceStep =
   StateM (\ s -> let instrCount' = (instrCount s) + 1
		  in  if (traceMarker s && instrCount' >= (traceNum s)) then
			 do 
			    putStrLn $ "\nNext is the " ++ 
				       show instrCount' ++ 
				       ". instruction (pc = " ++ 
				       show (pc s) ++ 
				       ")"
			    traceCode (code s) (pc s)
			    putStrLn "Options:"
			    putStrLn "1.    Inspect state"
			    putStrLn "2.    Fast proceed"
			    putStrLn "Else: provide step"
			    putStr   "-> "
			    answer <- getInt
			    putChar '\n'
			    case answer of
				 1  -> do trace s
					  return ((), s{ instrCount = instrCount' })
				 2  -> do putStr "Goto instruction -> "
					  traceNum' <- getInt
					  return ((), s{ instrCount = instrCount', traceNum = traceNum'})
				 _  -> return ((), s{ instrCount = instrCount' })
		      else
			 return ((), s{ instrCount = instrCount' }))

output string =
   StateM (\ s -> 
	     do putStrLn string
		return ((), s))


traceState =
   StateM (\ s -> do trace s
		     return ((), s))


traceInstrCount =
   StateM (\ s -> 
	     if (traceMarker s) then
		do putStrLn $ "There were " ++ show (instrCount s) ++ " instructions executed." 
		   return ((), s)
	      else
		return ((), s))

answer =
   StateM (\ s -> 
	     let h                    = heap s
		 (mainId, searchVars) = stab s
		 showResult =
		    do let e = env  s
		       putStrLn "-------- answer expression --------"
		       showTerm mainId e
		       putChar '\n'
		       mapM_ (\ (v, name) -> do putStr "\t"
						putStr $ name ++ " = " 
						showTerm v e
						putChar '\n') searchVars
		       putStrLn "-----------------------------------"

		 showTerm v e =
		    let p@(HeapAddr a) = lookupEnv e v
			b	       = lookupHeap' h p
		    in  case b of
			     Unbound -> let i = findIndex ((==) a) (map fst searchVars)
					in  case i of
						 Nothing -> putStr $ "_" ++ show a ++ " "
						 Just i' -> putStr $ snd (searchVars!!i')
			     Bound c -> showClosure c
			     _	     -> error "showTerm -> blackholemarker"

		 showClosure (Closure addr e) =
		    let instr       = lookupCode (code s) addr
		    in  case instr of
			     CONSTANT name ->    putStr $ name
			     CONSTR   name -> do putStr $ "(" ++ name ++ " "
						 let (INT arity) = lookupCode (code s) (addr + cArityOffset)
						 showArgs arity (addr + cPatternOffset) e
						 putStr ")"
			     UVAR	   -> let (INT x) = lookupCode (code s) (addr + 1)
					      in  showTerm x e
			     DEREF	   -> let (INT x) = lookupCode (code s) (addr + 1)
					      in  showTerm x e
			     TAKE  	   -> putStr "<<function>>"
			     i		   -> error $ "showClosure: unexpected instruction in closure!\n" ++ (show i)

		 showArgs 1 addr e = let (INT x) = lookupCode (code s) (addr + 1)
				     in  showTerm x e
		 showArgs a addr e = let (INT x) = lookupCode (code s) (addr + 1)
				     in  do showTerm x e
					    putChar ' '
					    showArgs (a-1) (addr+4) e


	     in	 do showResult
		    --trace state
		    if isEmptyStack (chpS s) then
		       do putStrLn "no (more) solutions!"
			  return (False, s)
		     else
		       do putStrLn "more? (y/n)"
			  putStr "-> "
			  a <- getLine
			  if (a == "y" || a == "Y") then
			     return (True, s)
			   else
			     return (False, s))


getPC =
   StateM (\ s ->
	     return ((pc s), s))


incPC =
   StateM (\ s ->
	     let pc' = (pc s) + 1
	     in  return ((), s{ pc = pc'}))

incPCBy n =
   StateM (\ s ->
	     let pc' = (pc s) + n
	     in  return ((), s{ pc = pc'}))

setPC addr =
   StateM (\ s ->
	     return ((), s{ pc = addr}))


nextInstr =
   StateM (\ s ->
	     return (lookupCode (code s) ((pc s) + 1), s))

instrAt addr =
   StateM (\ s ->
	     return (lookupCode (code s) addr, s))

instrAtPC =
   StateM (\ s ->
	     return (lookupCode (code s) (pc s), s))

allocateFrame =
   StateM (\ s ->
	     let (heap', p) = allocateFrame' (heap s)
	     in  return (p, s{ heap = heap' }))


allocateFrames n =
   StateM (\ s ->
	     let (heap',ps) = allocateFrames' (heap s) n
	     in  return (ps, s{ heap = heap' }))


unbound p =
   StateM (\ s ->
	     let heap' = unbound' (heap s) p
	     in  return ((), s{ heap = heap' }))


bind p closure =
   StateM (\ s ->
	     let heap' = bind' (heap s) p closure
	     in  return ((), s{ heap = heap' }))


markEntered p =
   StateM (\ s ->
	     let heap' = markEntered' (heap s) p
	     in  return ((), s{ heap = heap' }))

lookupHeap p =
   StateM (\ s ->
	     let c = lookupHeap' (heap s) p
	     in  return (c, s))


extendEnvironment p =
   StateM (\ s -> 
	     let env' = extendEnvironment' (env s) p
	     in  return ((), s{ env = env' }))

bindEnv x p =
   StateM  (\ s ->
	      let env' = bindEnv' (env s) x p
	      in  return ((), s{ env = env' }))

lookupCurEnv x =
   StateM (\ s -> 
	     let p = lookupEnv (env s) x
	     in  return (p, s))



makeClosure addr =
   StateM (\ s -> 
	     return (Closure addr (env s), s))



setClosure (Closure addr env') =
   StateM (\ s -> 
	     return ((), s{ pc = addr, env = env' }))


newThreads_D c1 c2 =
   StateM (\ s -> 
	     let dispatcher' = newThreads (dispatcher s) c1 c2
	     in return ((), s{ dispatcher = dispatcher' }))

newNotRunnableThreads_D =
   StateM (\ s -> 
	     let dispatcher' = newNotRunnableThreads (dispatcher s)
	     in return ((), s{ dispatcher = dispatcher' }))


actualizeThread_D addr =
   StateM (\ s -> 
	     let dispatcher' = actualizeThread (dispatcher s) (Thread (Closure addr (env s)) (v s) (a s) (r s))
	     in return ((), s { dispatcher = dispatcher' }))



nextThread_D =
   StateM (\ s -> 
	     let (dispatcher', Thread (Closure pc' env') v' a' r') = 
		    nextThread (dispatcher s)
	     in return ((),    s { pc         = pc'
			         , dispatcher = dispatcher'
				 , env	      = env'
				 , v	      = v'
				 , a	      = a'
				 , r	      = r' }))


childsTerminated_D =
   StateM (\ s -> 
	     return (childsTerminated (dispatcher s), s))


removeChilds_D =
   StateM (\ s -> 
	     let dispatcher' = removeChilds (dispatcher s)
	     in return ((), s{ dispatcher = dispatcher' }))


actualizeThreadsValue_D =
   StateM (\ s -> 
	     let dispatcher' = actualizeThreadsValue (dispatcher s) (v s)
	     in return ((), s{ dispatcher = dispatcher' }))


lSonThread_D =
   StateM (\ s -> 
	     let (dispatcher', Thread (Closure pc' env') v' a' r') = lSonThread (dispatcher s)
	     in return ((),   s { pc          = pc'
				, dispatcher  = dispatcher'
				, env	      = env'
				, v	      = v'
				, a	      = a'
				, r	      = r' }))

rSonThread_D =
   StateM (\ s -> 
	     let (dispatcher', Thread (Closure pc' env') v' a' r') = rSonThread (dispatcher s)
	     in return ((),   s { pc          = pc'
				, dispatcher  = dispatcher'
				, env	      = env'
				, v	      = v'
				, a	      = a'
				, r	      = r' }))


parentThread_D =
   StateM (\ s -> 
	     let (dispatcher', Thread (Closure pc' env') v' a' r') = parentThread (dispatcher s)
	     in return ((),   s { pc          = pc'
				, dispatcher  = dispatcher'
				, env	      = env'
				, v	      = v'
				, a	      = a'
				, r	      = r' }))


firstChildTerminated_D =
   StateM (\ s -> 
	     let dispatcher' = firstChildTerminated (dispatcher s)
	     in  return ((), s{ dispatcher = dispatcher' }))


childValues_D =
   StateM (\ s -> 
	     return (childValues (dispatcher s), s))

hat closure =
   StateM (\ s -> 
	     return (hat' (code s) (heap s) closure, s))

parentThreadExists_D =
   StateM (\ s -> 
	     return (parentThreadExists (dispatcher s), s))

runnableThreadsLeft_D =
   StateM (\ s -> 
	     return (runnableThreadsLeft (dispatcher s), s))

suspendThread_D p =
   StateM (\ s -> 
	     let dispatcher' = suspendThread (dispatcher s) p
	     in	 return ((), s{ dispatcher = dispatcher' }))


wakeUpThreads_D p =
   StateM (\ s -> 
	     let dispatcher' = wakeUpThreads (dispatcher s) p
	     in	 return ((), s{ dispatcher = dispatcher' }))


pushChoicePoint addr =
   StateM (\ s -> 
	     let chpS' = pushChoicePoint' s addr
	     in	 return ((), s {chpS = chpS'} ))

furtherChoicepoints =
   StateM (\ s ->
	     return (not (isEmptyStack (chpS s)),s))

backtrack =
   StateM (\ s -> 
	     return ((), backtrack' s))



setValue v' =
   StateM (\ s -> 
	     return ((), s{ v = v' }))



getValue =
   StateM (\ s -> 
	     return ((v s),s))




pushArgument p =
   StateM (\ s -> 
	     let a' = push (a s) p
	     in	 return ((), s{ a = a' }))


popArgument =
   StateM (\ s -> 
	     let a' = pop (a s)
	     in	 return (top (a s), s{ a = a' }))

argStackEmpty =
   StateM (\ s -> 
	     return (isEmptyStack (a s),s))



pushReturnStack closure =
   StateM (\ s -> 
	     let r' = push (r s) closure
	     in  return ((), s{ r = r' }))


popReturnStack =
   StateM (\ s -> 
	     let r' = pop (r s)
	     in  return (top (r s), s{ r = r' }))

returnStackEmpty =
   StateM (\ s -> 
	     return (isEmptyStack (r s),s))
