{ -- -*- haskell -*-
module Parser(parse)
where

import Scanner(Token(..))
} 


%name parse
%tokentype { Token }
%token  "("      { TokenLParen         }
        ")"	 { TokenRParen	       }
        "{"	 { TokenLCBracket      }
        "}"	 { TokenRCBracket      }
        "["	 { TokenLSBracket      }
        "]"	 { TokenRSBracket      }
	";"	 { TokenSemicolon      }
	":"	 { TokenColon	       }
	","	 { TokenComa	       }
        "="	 { TokenEqu	       }    
	"->"	 { TokenTo	       }     
	"lambda" { TokenLambda	       }    
	"let"    { TokenLet	       }
	"where"	 { TokenWhere	       }
	"free"   { TokenFree	       }
	"in"     { TokenIn	       }
--	"flex"	 { TokenFlex	       }
--	"rigid"	 { TokenRigid	       }
	"try"	 { TokenTry	       }
	"main"	 { TokenMain	       }
--	"eval"	 { TokenEval	       }
	"if"	 { TokenIf	       }
	"then"	 { TokenThen	       }
	"else"	 { TokenElse	       }
	"#"	 { TokenGuard	       }
	"&"	 { TokenAnd	       }
	"=:="	 { TokenUnify	       }
	C	 { TokenConstructor $$ }
	V	 { TokenVar $$	       }	  
%%


Program :: { Program }
    : "main" "=" Expr ";" FunDefs           { Program      $3 $5       }

FunDefs  :: { Functions }
    :  				            { [] }
    | FunDef ";" FunDefs		    { $1 : $3 }

FunDef :: { Function }
    : V Patterns "=" Expr WhereDef 	    { FunctionE (Var $1) $2 $4 $5 }
    | V Patterns GExprs   WhereDef	    { FunctionG (Var $1) $2 $3 $4 }

Patterns :: { Patterns }
    :					    { [] }
    | Pattern Patterns			    { $1 : $2 }

Pattern :: { Pattern }
    : V					    { PVar (Var $1)    }
    | CPattern				    { $1 }

CPattern :: { Pattern }
    : "(" C Patterns ")"		    { PConstr $2 $3 }
    | C 				    { PConstr $1 [] }
    | "[" Patterns "]"			    { PList   $2    }
    | "(" Pattern ":" Pattern ")"	    { PCons   $2 $4 }

WhereDef :: { LocalDefs }
    :					    { [] }
    | "where" "{" LocalDefs "}"		    { $3 }

LocalDefs :: { LocalDefs }
    : LocalDef				    { [$1] }
    | LocalDef ";" LocalDefs		    { $1 : $3 }

LocalDef  :: { LocalDef }
    : FunDef				    { LocalFun   $1 }
    | Patterndecl			    { LocalPat   $1 }
    | FreeDecl				    { LocalFree  $1 }

Patterndecl :: { PatternDecl }
    : CPattern "=" Expr WhereDef		    { PatternDecl $1 $3 $4 }

FreeDecl :: { FreeDecl }
    : Vars "free"			    { Free $1 }

GExprs :: { GExprs }
    : GExpr                                 { [$1] }
    | GExpr GExprs			    { $1 : $2 }

GExpr :: { GExpr }
    : "#" Expr "=" Expr			    { GExpr $2 $4 }

Exprs :: { Exprs }
    :					    { [] }
    | Exprs2				    { $1 }


Exprs2 :: { Exprs }
    : Expr				    { [$1] }
    | Expr "," Exprs2			    { $1 : $3	  }

Expr :: { Expr }
    : "let" "{" LocalDefs "}" "in" Expr     { LetExpr      $3 $6       }
    | "lambda" Patterns "->" Expr	    { LambdaExpr   $2 $4       }
    | "if" Expr "then" Expr "else" Expr	    { IfExpr	   $2 $4 $6    }
    | "try" Expr			    { TryExpr	   $2         }
    | FunctExpr				    { $1 }
    | "(" InfixExpr ")"			    { $2 }

FunctExpr :: { Expr }
    : BasicExpr				    { $1 }
    | FunctExpr BasicExpr		    { AppExpr $1 $2 }

BasicExpr :: { Expr }
    : V					    { VarExpr	   (Var $1)    }
    | C					    { ConstrExpr   $1 []       }
    | "[" Exprs "]"			    { ListExpr	   $2	       }
    | "(" Expr ")"			    { $2 }

InfixExpr :: { Expr }
    : Expr "&" AndExpr			    { AndExpr   $1 $3 }
    | Expr "=:=" Expr			    { UnifyExpr $1 $3 }
    | Expr ":" ConsExpr			    { ConsExpr  $1 $3 }

AndExpr :: { Expr }
    : Expr				    { $1 }
    | Expr "&" AndExpr			    { AndExpr $1 $3 }

ConsExpr :: { Expr }
    : Expr				    { $1 }
    | Expr ":" ConsExpr			    { ConsExpr $1 $3 }

Vars :: { Vars }
    : V					    { [Var $1] }
    | V "," Vars			    { (Var $1) : $3 }


{
data Program = Program Expr Functions
     deriving (Show)

type Functions = [Function]

data Function = FunctionE Var Patterns Expr   LocalDefs
	      | FunctionG Var Patterns GExprs LocalDefs
     deriving (Show)

type Patterns = [Pattern]

data Pattern  = PVar  Var
	      | PConstr String Patterns
	      | PList   Patterns
	      | PCons	Pattern Pattern
     deriving (Show)

type LocalDefs = [LocalDef]

data LocalDef = --LocalAnnot Annot
	        LocalFun   Function
	      | LocalPat   PatternDecl
	      | LocalFree  FreeDecl
     deriving (Show)

data PatternDecl = PatternDecl Pattern Expr LocalDefs
     deriving (Show)

data FreeDecl = Free Vars
     deriving (Show)

type Vars = [Var]

data Var    = Var    String
     deriving (Show)

type GExprs = [GExpr]

data GExpr = GExpr Expr Expr
     deriving (Show)

type Exprs = [Expr]	        

data Expr = LetExpr        LocalDefs   Expr
	  | LambdaExpr     Patterns    Expr
	  | IfExpr	   Expr	       Expr	Expr
	  | TryExpr	   Expr
	  | AppExpr	   Expr	       Expr
	  | ConstrExpr	   String      Exprs
	  | ListExpr	   Exprs
	  | AndExpr	   Expr	       Expr
	  | UnifyExpr	   Expr	       Expr
	  | ConsExpr	   Expr	       Expr
	  | VarExpr	   Var
     deriving (Show)


happyError t = error $ "Fehler nahe " ++ (show (take 5 t)) ++ "\n"

}
